#define _BSD_SOURCE //for random() and srandom()
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "iterator.h"
#include "graph.h"
#include "links.h"
#include "nodes.h"
#include "graph-algo.h"



struct bst_node** gen_destinations (graph* g, long nb_dest) {
  /* Choose destinations */
  struct bst_node **dest_set;
  node_id cur_nb_d, d;
  assert( (dest_set = (struct bst_node **)malloc(sizeof(struct bst_node*))) != NULL);
  *dest_set=NULL;

  cur_nb_d=0;
  while( cur_nb_d < nb_dest ) {
    d=random()%g->n;
    cur_nb_d += insert_count(dest_set,d);
  }

  return (dest_set);
}

node_id gen_source (graph* g) {  
  return (random()%g->n);
}


/* returns a node_list containing the neighbours of u in g */
/* in a random order */
node_list *neighbours_r(graph *g, node_id u){
  node_list* rlist;
  node_id nb_nodes,i,j,tmp;
  queue *q;
  struct bst_node *mybstnode;

  assert( u < g->n );
  assert( (rlist = (node_list *)malloc(sizeof(node_list))) != NULL);
  assert( (rlist->nodes = (node_id *)malloc(g->n*sizeof(node_id))) != NULL );
  nb_nodes = 0;

  q = iterate_init(g->adjacencies[u],g->n);
  mybstnode = iterate_next_node(q);
  while( mybstnode != NULL ){
    rlist->nodes[nb_nodes] = mybstnode->data;
    nb_nodes ++;
    mybstnode = iterate_next_node(q);
  }
  iterate_end(q);
  rlist->nb_nodes = nb_nodes;

  /* Random permutation of nodes in rlist */
  for(i=nb_nodes-1;i>=0;i--){
    j = random()%(i+1);
    tmp = rlist->nodes[i];
    rlist->nodes[i] = rlist->nodes[j];
    rlist->nodes[j] = tmp;
  }
  return(rlist);
}

node_id *bfs_r(graph *g, node_id v){
  node_id *tree;
  node_id i,u,neigh;
  node_id_queue *qbfs;
  node_list *rlist;

  assert( (tree = (node_id *)malloc(g->n*sizeof(node_id))) != NULL );
  for(i=0;i<g->n;i++)
    tree[i] = -1;
  
  qbfs = empty_node_id_queue(g->n);
  node_id_queue_add(qbfs,v);
  tree[v] = v;

  while( !is_empty_node_id_queue(qbfs) ){
    u = node_id_queue_get(qbfs);
    rlist = neighbours_r(g,u);
    for(i=0;i<rlist->nb_nodes;i++){
      neigh = rlist->nodes[i];
      if( tree[neigh] == -1 ){
	tree[neigh] = u;
	node_id_queue_add(qbfs,neigh);	
      } 
    }
    free_node_list(rlist);
  }
  free_node_id_queue(qbfs);
  return(tree);
}

node_id *bfs_deterministic(graph *g, node_id v){
  node_id *tree;
  node_id i,u,neigh;
  node_id_queue *qbfs;
  queue  *qiter;
  struct bst_node *mybstnode;

  assert( (tree = (node_id *)malloc(g->n*sizeof(node_id))) != NULL );
  for(i=0;i<g->n;i++)
    tree[i] = -1;
  
  qbfs = empty_node_id_queue(g->n);
  node_id_queue_add(qbfs,v);
  tree[v] = v;

  while( !is_empty_node_id_queue(qbfs) ){
    u = node_id_queue_get(qbfs);

    qiter = iterate_init(g->adjacencies[u],g->n);
    mybstnode = iterate_next_node(qiter);
    while( mybstnode != NULL ){
      neigh = mybstnode->data;
      if( tree[neigh] == -1 ){
	tree[neigh] = u;
	node_id_queue_add(qbfs,neigh);	
      }
      mybstnode = iterate_next_node(qiter);
    }
    iterate_end(qiter);
  }
  free_node_id_queue(qbfs);
  return(tree);
}

node_id *bfs_dist(graph *g, node_id v){
  node_id *dist, d, i, u, neigh;
  queue  *qiter;
  node_id_queue *qbfs;
  struct bst_node *mybstnode;

  assert( (dist = (node_id *)malloc(g->n*sizeof(node_id))) != NULL );
  for(i=0;i<g->n;i++)
    dist[i] = -1;

  qbfs = empty_node_id_queue(g->n);
  node_id_queue_add(qbfs,v);
  dist[v] = 0;

  while( !is_empty_node_id_queue(qbfs) ){
    u = node_id_queue_get(qbfs);
    d = dist[u];
    qiter = iterate_init(g->adjacencies[u],g->n);
    mybstnode = iterate_next_node(qiter);
    while( mybstnode != NULL ){
      neigh = mybstnode->data;
      if( dist[neigh] == -1 ){
	dist[neigh] = d+1;
	node_id_queue_add(qbfs,neigh);	
      } 
      mybstnode = iterate_next_node(qiter);
    }
    iterate_end(qiter);
  }
  free_node_id_queue(qbfs);
  return(dist);
}


/* tree is a bfs */
/* Creates a set of the nodes that are on the path from the source */
/* to any of the nodes in dest_set. */
/* nb_dest is needed for the size of the queue to iterate over dest_set */
struct bst_node **partial_bfs(node_id *tree, 
                              struct bst_node **dest_set, 
                              node_id nb_dest,
                              struct link **links_seen, 
                              long *nb_links_seen_part) {
  struct bst_node **nodes_seen;
  struct bst_node *dest_node;
  node_id nb_nodes_seen, cur_dest, u, new_node;
  long nb_links = 0;
  queue *q;

  assert( (nodes_seen = (struct bst_node **)malloc(sizeof(struct bst_node*))) != NULL);
  *nodes_seen = NULL;
  nb_nodes_seen = 0;
  q = iterate_init(*dest_set,nb_dest);
  dest_node = iterate_next_node(q); 
  
  while( dest_node != NULL ){ 
    cur_dest = dest_node->data;
    if( tree[cur_dest] != -1 ){
      u = cur_dest;
      new_node = insert_count(nodes_seen, u);
      while( new_node == 1 ){
	nb_nodes_seen += new_node;
	/* printf("Inserted node %ld\n",u); */
        if (u != tree[u]) {
          nb_links += insert_link_count (links_seen, u, tree[u]);
        }
	u = tree[u];
	new_node = insert_count(nodes_seen,u);
      }
    }
    dest_node = iterate_next_node(q);
  } 
  /* printf("nb nodes seen: %ld\n",nb_nodes_seen); */
  iterate_end(q); 
  /* printf("end of iteration on dests\n"); */

  *nb_links_seen_part = nb_links;
  return(nodes_seen);
}

node_id bfs_from_dest(graph *g, node_id d, node_id *dist,
                      struct bst_node **nodes_seen, 
                      struct link** links_seen, long *nb_links_part){
  queue  *qiter;
  node_id_queue *qbfs;
  node_id u, du, neigh, nb_seen;
  struct bst_node *mybstnode;

  /* Check that destination is reachable from source */
  if( dist[d] == -1 ){
    return(0);
  }

  nb_seen = 0;

  qbfs = empty_node_id_queue(g->n);  
  node_id_queue_add(qbfs,d);
  nb_seen += insert_count(nodes_seen,d);

  while( ! is_empty_node_id_queue(qbfs) ){
    u = node_id_queue_get(qbfs);
    du = dist[u];

    qiter = iterate_init(g->adjacencies[u],g->n);
    mybstnode = iterate_next_node(qiter);
    while( mybstnode != NULL ){
      neigh = mybstnode->data;
      if( dist[neigh] < du ){
        *nb_links_part += insert_link_count (links_seen, u, neigh);
	if( insert_count(nodes_seen, neigh) == 1 ){
	  nb_seen += 1;
	  node_id_queue_add(qbfs,neigh);
	}
      }
      mybstnode = iterate_next_node(qiter);
    }
    iterate_end(qiter);
  }
  free_node_id_queue(qbfs);
  return(nb_seen);
}

struct bst_node **asp(graph *g, node_id s, 
                      struct bst_node **dest_set, long nb_dest,
                      struct link **links_seen, 
                      long* nb_asp_nodes, long* nb_links){
  struct bst_node **nodes_seen, *dest_node;
  node_id *dist, cur_dest, nb_seen;
  long nb_links_seen;
  queue *q;

  assert( (nodes_seen = (struct bst_node **)malloc(sizeof(struct bst_node*))) != NULL);
  *nodes_seen = NULL;
  dist = bfs_dist(g,s);
  nb_seen = 0;
  nb_links_seen = 0;


  q = iterate_init(*dest_set,nb_dest);
  dest_node = iterate_next_node(q); 
  while( dest_node != NULL ){
    cur_dest = dest_node->data;
    nb_seen += bfs_from_dest(g, cur_dest, dist,
                             nodes_seen, 
                             links_seen,
                             &nb_links_seen);

    dest_node = iterate_next_node(q);
  } 
  iterate_end(q); 
  free(dist);

  *nb_asp_nodes = nb_seen;
  *nb_links = nb_links_seen;
  return(nodes_seen);
}


struct bst_node** usp(graph* g, node_id src, 
                      struct bst_node** dest_set, long nb_dest,
                      struct link** p_links_seen, long* nb_links_seen_part) {
  struct bst_node** p_nodes;
  node_id* tree;

  tree = bfs_r(g,src);
  p_nodes = partial_bfs(tree, dest_set, nb_dest,
                        p_links_seen, nb_links_seen_part);

  free (tree);
  return p_nodes;
}


struct bst_node** deterministic_usp(graph* g, node_id src, struct bst_node** dest_set, long nb_dest,
                                    struct link** p_links_seen, long* nb_links_seen_part) {
  struct bst_node** p_nodes;
  node_id* tree;

  tree = bfs_deterministic(g,src);

  p_nodes = partial_bfs(tree, dest_set, nb_dest,
                        p_links_seen, nb_links_seen_part);

  free (tree);
  return p_nodes;
}



void swap_link (graph* g) {
  long link1, link2;
  node_id a1, b1, a2, b2;

  /*
    a1 --- b1     link1
    a2 --- b2     link2
   */
  link1 = random()%g->m;
  a1 = g->linklistleft[link1];
  b1 = g->linklistright[link1];

  link2 = random()%g->m;
  a2 = g->linklistleft[link2];
  b2 = g->linklistright[link2];
  
  // if links a1 -- b2 or b1 -- a2 exist we should not make swap.
  // also we should not make any swaps, if a1 == a2 or b1 == a2
  while (
         ( (* search(g->adjacencies+a1, b2) ) != NULL )
         ||
         ( (* search(g->adjacencies+a2, b1) ) != NULL )
         ||
         a1 == b2
         ||
         a2 == b1
         ) {
      link2 = random()%g->m;
      a2 = g->linklistleft[link2];
      b2 = g->linklistright[link2];
  }

  /* remove links
     a1   b1
     a2   b2
   */
  remove_link (g, link1);
  remove_link (g, link2);


  /*
    a1 -----\     /---- b1
             \   /        
            -----
           /   \
    a2 ---/     \------ b2

  */

  add_link(g, link1, a1, b2);
  add_link(g, link2, a2, b1);

}

//TODO: prepare_for_link_swap
//TODO: first part of link-swap. Links removing.
//TODO: second part of link-swap. Links addtion.

void rewire (graph *g) {
  long link_id = remove_random_link (g);
  add_random_link (g, link_id);
}

// add a random link 'a--b'
void add_random_link (graph *g, long link_id) {
  int link_added = 0;
  int u, v;
  while (!link_added) {
    /* add one link */
    u = 0;
    v = 0;
    while (u == v) {
      u = random()%g->n;
      v = random()%g->n;
    }

    /* test if link already exists */
    link_added = add_link (g, link_id, u, v);
  }
}


long remove_random_link (graph *g) {
  long link;
  link = random()%g->m;
  remove_link(g, link);
  return (link);
}
