#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "iterator.h"
#include "links.h"


struct link* iterate_next_link (queue *q) {
  struct link* link;

  if( is_empty_queue(q) )
    return(NULL);
  link = (struct link*)queue_get(q);
  if( link->left != NULL )
    queue_add(q, (void *)link->left);
  if( link->right != NULL )
    queue_add(q, (void *)link->right);
  return(link);
}

//we add a link (a,b) in such manner that a <= b,
//if not, we swap 'a' with 'b'.
struct link* new_link(node_id a, node_id b) {
  struct link* my_link;
  assert((my_link = (struct link*)malloc(sizeof(struct link))) != NULL);
  if (a <= b) {
    my_link->a = a;
    my_link->b = b;
  } else {
    my_link->a = b;
    my_link->b = a;
  }
  my_link->left = NULL;
  my_link->right = NULL;

  return my_link;
}

struct link** search_link(struct link** links, node_id a, node_id b) {
  //do not forget about swap values, plz. bro. do it carefuly.
  struct link** l = links;

  //swap a <-> b
  node_id c;
  if (a > b) {
    c = b;
    b = a;
    a = c;
  }
  
  while ( (*l) != NULL) {
    if ( a < (*l)->a )
      l = &(*l)->left;
    else if ( a > (*l)->a )
      l = &(*l)->right;
    else { //when a == (*l)->a
      if ( b < (*l)->b )
        l = &(*l)->left;
      else if ( b > (*l)->b )
        l = &(*l)->right;
      else //when link is found
        break; 
    }
  }

  return l;
}

int insert_link_count(struct link** links, node_id a, node_id b) {
  //swap a <-> b
  node_id c;
  if (a > b) {
    c = b;
    b = a;
    a = c;
  }

  struct link** my_link = search_link(links, a, b);


  if (*my_link == NULL) {
    *my_link = new_link(a, b);
    return 1;
  }
  return 0;
}


// Computes the union of p_links and links_seen 
// Union will be stored in links_seen and nb_links_seen will be changed.
void links_union(struct link** links_seen, long* nb_links_seen, 
                 struct link** p_links_seen, long nb_links_seen_part) {

  queue *q;
  struct link *link;
  q = iterate_init(*p_links_seen, nb_links_seen_part);
  link = iterate_next_link(q);
  while( link != NULL ) {
    *nb_links_seen += insert_link_count(links_seen, link->a, link->b);
    link = iterate_next_link(q); 
  } 
  iterate_end(q);

}



void clear_link_set (struct link** set1, node_id size1) {
  queue *q1;
  struct link *link;

  q1 = iterate_init(*set1, size1);
  link = iterate_next_link(q1);
  while( link != NULL ){
    free(link);
    link = iterate_next_link(q1);
  }
  iterate_end(q1);
  *set1 = NULL;
}

void free_link_set(struct link** set1, node_id size1){
  clear_link_set(set1, size1);
  free(set1);
}


void save_links(struct link** links, long nb_links, FILE* file) {
  queue *q1;
  struct link *link;

  q1 = iterate_init(*links, nb_links);
  link = iterate_next_link(q1);
  while( link != NULL ){
    fprintf(file, "%ld %ld\n", link->a, link->b);
    link = iterate_next_link(q1);
  }
  iterate_end(q1);
}
