#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "iterator.h"
#include "nodes.h"


struct bst_node *iterate_next_node (queue *q) {
  struct bst_node *node;

  if( is_empty_queue(q) )
    return(NULL);
  node = (struct bst_node *)queue_get(q);
  if( node->left != NULL )
    queue_add(q, (void *)node->left);
  if( node->right != NULL )
    queue_add(q, (void *)node->right);
  return(node);
}

struct bst_node* new_bst_node(node_id data) {
  struct bst_node* my_node;

  assert( (my_node = (struct bst_node *)malloc(sizeof(struct bst_node))) != NULL);
  my_node->data = data;
  my_node->left = NULL;
  my_node->right = NULL;
  return my_node;
}


struct bst_node** search(struct bst_node** root, node_id data) {
  struct bst_node** node = root;

  while (*node != NULL) {
    if (data < (*node)->data)
      node = &(*node)->left;
    else if (data > (*node)->data)
      node = &(*node)->right;
    else //when node is found.
      break;
  }
  return node;
}

void insert(struct bst_node** root, node_id data) {
  struct bst_node** node = search(root, data);
  if (*node == NULL) {
    *node = new_bst_node(data);
  }
}

/* same as insert but returns 0 if data was already in bst, */
/* 1 else */
int insert_count(struct bst_node** root, node_id data) {
  struct bst_node** node = search(root, data);
  if (*node == NULL) {
    *node = new_bst_node(data);
    return(1);
  }
  return(0);
}


void delete(struct bst_node** node) {
  struct bst_node* old_node = *node;
  if ((*node)->left == NULL) {
    *node = (*node)->right;
    free (old_node);
  } else if ((*node)->right == NULL) {
    *node = (*node)->left;
    free (old_node);
  } else {
    struct bst_node** pred = &(*node)->left;
    while ((*pred)->right != NULL) {
      pred = &(*pred)->right;
    }

    /* Swap values */
    node_id temp = (*pred)->data;
    (*pred)->data = (*node)->data;
    (*node)->data = temp;

    delete(pred);
  }
}


// Computes the union of p_nodes and nodes_seen 
// Union will be stored in nodes_seen and nb_nodes_seen will be changed.
void nodes_union(struct bst_node** nodes_seen, long* nb_nodes_seen,
                 struct bst_node** p_nodes, long nb_p_nodes) {
  queue *q;
  struct bst_node *node;
  q = iterate_init(*p_nodes, nb_p_nodes);
  node = iterate_next_node(q);
  while( node != NULL ) {
    *nb_nodes_seen += insert_count(nodes_seen, node->data);
    node = iterate_next_node(q); 
  } 
  iterate_end(q);
}

//ATTENTION: it's a hack-trap.
void print_nodes(struct bst_node** p_nodes, long nb_p_nodes) {
  queue *q;
  struct bst_node *node;
  q = iterate_init(*p_nodes, nb_p_nodes);
  node = iterate_next_node(q);
  while( node != NULL ) {
    printf("%ld ",node->data);
    node = iterate_next_node(q); 
  } 
  iterate_end(q);
}



void free_node_list(node_list *rlist){
  assert( rlist != NULL);
  assert( rlist->nodes != NULL);
  free(rlist->nodes);
  free(rlist);
}



/* queue for node_id */
// FIXME: remove\or rewrite this trash. i'm should use queue.{c,h} instead.

node_id_queue *empty_node_id_queue(node_id size){
  node_id_queue *q;

  assert( (q=(node_id_queue *)malloc(sizeof(node_id_queue))) != NULL );
  q->size = size+1;
  assert( (q->elts = (node_id *)malloc((size+1)*sizeof(node_id))) != NULL );
  q->begin = q->end = 0;
  return(q);
}

void free_node_id_queue(node_id_queue *q){
  assert( (q != NULL) && (q->elts != NULL) );
  free(q->elts);
  free(q);
}

int is_empty_node_id_queue(node_id_queue *q){
  return(q->begin == q->end);
}

int is_full_node_id_queue(node_id_queue *q){
  return(q->begin == (q->end+1)%q->size);
}

void node_id_queue_add(node_id_queue *q, node_id e){
  assert(!is_full_node_id_queue(q));
  assert( q->end < q-> size );
  q->elts[q->end] = e;
  q->end++;
  q->end %= q->size;
}

node_id node_id_queue_get(node_id_queue *q){
  node_id r;

  r = q->elts[q->begin];
  q->begin++;
  q->begin %= q->size;
  return(r);
}



//FIXME: 
void free_set(struct bst_node **set1, node_id size1) {
  queue *q1;
  struct bst_node *n1;

  q1 = iterate_init(*set1, size1);
  n1 = iterate_next_node(q1);
  while( n1 != NULL ){
    free (n1);
    n1 = iterate_next_node(q1);
  }
  iterate_end(q1);
  
  free(set1);
}
