#define _BSD_SOURCE //for random() and srandom()
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <unistd.h>
#include <malloc.h>
//#include <math.h>
#include <time.h>
#include <string.h>


#include "iterator.h"
#include "links.h"
#include "nodes.h"
#include "graph.h"
#include "graph-algo.h"
#include "er-graph.h"
#include "pl-graph.h"

void run_experiment(graph* g, long nb_dest, long nb_swaps, long nb_steps,
                    char* outdir) {
  struct bst_node** dest_set;
  struct bst_node** asp_nodes;
  struct bst_node** usp_nodes;

  struct bst_node** usp_nodes_seen;
  struct bst_node** asp_nodes_seen;

  struct link** usp_links;
  struct link** asp_links;

  struct link** usp_links_seen;
  struct link** asp_links_seen;

  node_id src;

  long nb_usp_links_part = 0;
  long nb_usp_nodes_part = 0;
  long cur_step;
  long nb_asp_nodes_part, nb_asp_links_part;
  long cum_nb_usp_nodes, cum_nb_usp_links = 0;
  long cum_nb_asp_nodes, cum_nb_asp_links = 0;

  dest_set = gen_destinations(g, nb_dest);
  src = gen_source(g);

  assert( (usp_nodes_seen = (struct bst_node **)malloc(sizeof(struct bst_node*))) != NULL );
  *usp_nodes_seen = NULL;
  cum_nb_usp_nodes = 0;

  assert( (asp_nodes_seen = (struct bst_node **)malloc(sizeof(struct bst_node*))) != NULL );
  *asp_nodes_seen = NULL;
  cum_nb_asp_nodes = 0;


  assert( (usp_links = (struct link **)malloc(sizeof(struct link*))) != NULL );
  *usp_links = NULL;

  assert( (asp_links = (struct link **)malloc(sizeof(struct link*))) != NULL );
  *asp_links = NULL;

  assert( (usp_links_seen = (struct link **)malloc(sizeof(struct link*))) != NULL );
  *usp_links_seen = NULL;
  cum_nb_usp_links = 0;

  assert( (asp_links_seen = (struct link **)malloc(sizeof(struct link*))) != NULL );
  *asp_links_seen = NULL;
  cum_nb_asp_links = 0;


  //perform some measurements
  for (cur_step = 0; cur_step < nb_steps; cur_step ++) {

    nb_usp_nodes_part = 0; 
    nb_usp_links_part = 0; 
    nb_asp_nodes_part = 0; 
    nb_asp_links_part = 0; 

    usp_nodes = usp(g, src, dest_set, nb_dest, usp_links, 
                    &nb_usp_links_part);

    nb_usp_nodes_part = nb_usp_links_part + 1;

    asp_nodes = asp(g, src, dest_set, nb_dest, 
                    asp_links, 
                    &nb_asp_nodes_part, &nb_asp_links_part);


    //usp
    nodes_union(usp_nodes_seen, &cum_nb_usp_nodes, 
                usp_nodes, nb_usp_nodes_part);
    links_union(usp_links_seen, &cum_nb_usp_links,
                usp_links, nb_usp_links_part);
    free_set(usp_nodes, nb_usp_nodes_part);


    //asp
    nodes_union(asp_nodes_seen, &cum_nb_asp_nodes, asp_nodes, nb_asp_nodes_part);
    links_union(asp_links_seen, &cum_nb_asp_links,
                asp_links, nb_asp_links_part);
    free_set(asp_nodes,g->n);
    clear_link_set (asp_links, nb_asp_links_part);

    
    //print them all!
    printf("%ld %ld %ld %ld %ld\n", 
           cur_step, cum_nb_usp_nodes, cum_nb_usp_links,
           cum_nb_asp_nodes, cum_nb_asp_links);

    ///save path
    char name[256];
    snprintf (name, sizeof name, "%s/%ld.path", outdir, cur_step);
    FILE* file = fopen(name, "w");
    save_links (usp_links, nb_usp_links_part, file);
    fclose(file);
    //\save path
    clear_link_set (usp_links, nb_usp_links_part);
    
    //do swaps!
    for (int i = 0; i < nb_swaps; i++) {
      swap_link(g);
    }
    
  }

  free_link_set(asp_links, nb_asp_links_part);
  free_link_set(usp_links, nb_usp_links_part);
  free_set(usp_nodes_seen, cum_nb_usp_nodes);
  free_link_set(usp_links_seen, cum_nb_usp_links);
  free_set(asp_nodes_seen, cum_nb_asp_nodes);
  free_link_set(asp_links_seen, cum_nb_asp_links);
  free_set(dest_set,nb_dest);
}

void main_pl(int argc, char** argv) {
  long n,m,nb_dest,nb_step,nb_swaps;
  double p;

  if( argc != 7 ){
    fprintf(stderr, 
"Usage: %s nb_nodes exp max_links_for_that_exp nb_swaps nb_step outdir\n\
       Sorry for max_links_for_that_exp :) \n",argv[0]);
    exit(1);
  }

  sscanf(argv[1],"%ld",&n);
  sscanf(argv[2],"%lf",&p);
  sscanf(argv[3],"%ld",&m);
  nb_dest = 1;
  sscanf(argv[4],"%ld",&nb_swaps);
  sscanf(argv[5],"%ld",&nb_step);
  

  graph* g = gen_pl_graph(n, -p, m);
  run_experiment(g, nb_dest, nb_swaps, nb_step, argv[6]);
  free_graph(g);
}

void main_er(int argc, char** argv) {
  long n,m,nb_dest,nb_step,nb_swaps;

  if( argc != 6 ){
    fprintf(stderr, 
            "Usage: %s nb_nodes nb_links nb_swaps nb_step outdir\n",argv[0]);
    exit(1);
  }

  sscanf(argv[1],"%ld",&n);
  sscanf(argv[2],"%ld",&m);
  sscanf(argv[3],"%ld",&nb_swaps);
  sscanf(argv[4],"%ld",&nb_step);
  nb_dest = 1;

  graph* g = gen_er_graph(n, m);
  run_experiment(g, nb_dest, nb_swaps, nb_step, argv[5]);
  free_graph(g);
}

int main(int argc, char** argv) {
  //TODO: just use another seed here.
  srandom(time(NULL) + getpid());

  mallopt(M_TRIM_THRESHOLD,-1);
  mallopt(M_MMAP_MAX,0);

  if( argc < 2 ){
    fprintf(stderr, 
            "Usage: %s er|pl [bla..bla..bla]\n", argv[0]);
    exit(1);
  }
  
  if (strcmp(argv[1],"er") == 0) {
    main_er(argc-1, argv+1);
  } else if (strcmp(argv[1],"pl") == 0) {
    main_pl(argc-1, argv+1);
  } else {
    fprintf(stderr, 
            "Usage: %s er|pl [bla..bla..bla]\n",argv[0]);
    exit(1);
  }



  return (0);
}
