#define _BSD_SOURCE //for random() and srandom()
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "graph.h"
#include "pl-graph.h"


// thx to Guillaume and Latapy from http://complexnetworks.fr
// for the original version see     http://jlguillaume.free.fr/www/programs.php
// 
// Here 'pl' means 'Power law'

//some code for generate random numbers

// NOTE: u should use a negative p
// HELP!!!: i'm totally confused with formulae.
int random_pl (int m, int M, double p) {
    double r=(random()*1.)/(RAND_MAX*1.);
    double nexp=p+1.;
    double norm=1./(pow(M*1.0,nexp)-pow(m*1.0,nexp));
    double expo=log10(r/norm+pow(m*1.0, nexp))/nexp;
    return (int)floor(pow(10.,expo));
}

void swap(int* a, int* b) {
  int tmp=*a;
  *a=*b;
  *b=tmp;
}

// m and M are minimum and maximum values respectivelly
int add_random_links_pl(graph *g, int m, int M, double p) {
  int* numbers;
  long n = g->n;
  long sum = 0;

  // generate degree sequence with respect to power law
  assert( (numbers = (int*) malloc(n *sizeof(int))) != NULL);
  for(int i=0 ; i<n ; i++) {
    numbers[i]=random_pl(m, M, p);
    sum += numbers[i];
  }

  if (sum % 2 != 0) {
    //Oups! you certanly should recall that function,
    //caz' sum of degrees is not even even and we cannot generate PL graph
    free(numbers);
    return(0);
  }

  int* connections;
  long t = 0;
  assert( (connections = (int*) malloc(sum *sizeof(int))) != NULL);
  for(int i=0; i<n; i++) {
    for (int j = 0; j<numbers[i]; j++){
      connections[t]=i;
      t++;      
    }
  }

  long nb_links = 0;
  long i = 0;
  // Construct some links
  
  while (i < sum) {
    /* swap values */
    
    long u = 0;
    long v = 0;
    

    swap(& (connections[i]),
         & (connections[rand()%(sum-i)+i]) );
    swap(& (connections[i+1]), 
         & (connections[rand()%(sum-i-1)+i+1]));
    u = connections[i];
    v = connections[i+1];

    // Erased configuration model --> configuration model
    // under some conditions...
    //
    // http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.159.4790
    // "Generating simple random graphs with prescribed degree distribution (2005)" 
    //  by Tom Britton , Maria Deijfen , Anders Martin-Löf

    if (u == v) { //erase self-loops
      //printf("# self-loop erased\n");
      i += 2;
      continue;
    }
        
    nb_links += add_link (g, nb_links, u, v);
    i += 2;
  }
  
  printf("# links in gr (current, expected): %ld %ld\n", nb_links, i/2);
  // todo: realloc memory for... for linklist{right,left}, and...
  g->m = nb_links;
  
  // don't forget to free all previously allocated memory
  free(connections);
  free(numbers);
  return (1);
}

// u should specify a negative p
// otherwise the behaviour is undefined.
graph* gen_pl_graph(long n, double p, /*FIXME: REMOVEME*/ long maxlinks) {
  //WHAT maxlinks, dude? 
  graph* g = empty_graph(n,maxlinks);
  int f = 0;
  while (!f) {
    f = add_random_links_pl(g, 1, 1000000, p);
  }
  return (g);
}
