#define _BSD_SOURCE //for random() and srandom()
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <unistd.h>
#include <malloc.h>
#include <time.h>
#include <string.h>


#include "iterator.h"
#include "links.h"
#include "nodes.h"
#include "graph.h"
#include "graph-algo.h"
#include "er-graph.h"
#include "pl-graph.h"

void run_proba_experiment(graph* g, long nb_dest, long nb_swaps, long nb_steps) {
  node_id src;
  struct bst_node** dest_set;
  struct bst_node** asp_nodes;
  struct bst_node** usp_nodes;
  struct bst_node** asp_nodes2;
  struct bst_node** usp_nodes2;
  struct bst_node** usp_nodes2_r;
  struct bst_node** asp_nodes_seen;
  struct link** usp_links;
  struct link** asp_links;
  struct link** usp_links2;
  struct link** usp_links2_r;
  struct link** asp_links2;
  struct link** usp_links_seen;
  struct link** asp_links_seen;
  long nb_asp_nodes_part, nb_asp_links_part;
  long nb_asp_nodes_part2, nb_asp_links_part2;
  long nb_usp_links_part = 0;
  long nb_usp_links_part2 = 0;
  long nb_usp_links_part2_r = 0;
  long nb_usp_nodes_part = 0;
  long nb_usp_nodes_part2 = 0;
  long nb_usp_nodes_part2_r = 0;
  
  long usp_diff_nodes = 0;
  long usp_diff_links = 0;
  long usp_diff_nodes_r = 0;
  long usp_diff_links_r = 0;
  long asp_diff_nodes = 0;
  long asp_diff_links = 0;


  //ATTENTION: prepare the big enough array for temporary storing 
  //           links from 'bfs_partial'
  assert( (usp_links    = (struct link **)malloc(sizeof(struct link*))) != NULL);
  *usp_links = NULL;
  assert( (asp_links    = (struct link **)malloc(sizeof(struct link*))) != NULL);
  *asp_links = NULL;
  assert( (usp_links2   = (struct link **)malloc(sizeof(struct link*))) != NULL);
  *usp_links2 = NULL;
  assert( (usp_links2_r = (struct link **)malloc(sizeof(struct link*))) != NULL);
  *usp_links2_r = NULL;
  assert( (asp_links2   = (struct link **)malloc(sizeof(struct link*))) != NULL);
  *asp_links2 = NULL;

  assert( (usp_links_seen = (struct link **)malloc(sizeof(struct link*))) != NULL );
  *usp_links_seen = NULL;

  assert( (asp_links_seen = (struct link **)malloc(sizeof(struct link*))) != NULL );
  *asp_links_seen = NULL;

  assert( (asp_nodes_seen = (struct bst_node **)malloc(sizeof(struct bst_node*))) != NULL );
  *asp_nodes_seen = NULL;

  dest_set = gen_destinations(g, nb_dest);
  src = gen_source(g);


  // 1. compute asp
  //    compute deterministic usp
  asp_nodes = asp(g, src, dest_set, nb_dest, 
                  asp_links, 
                  &nb_asp_nodes_part, &nb_asp_links_part);
  usp_nodes = deterministic_usp(g, src, dest_set, nb_dest, usp_links, 
                                &nb_usp_links_part);

  nb_usp_nodes_part = nb_usp_links_part + 1;



  // 2. do a few swaps
  // ici.
  // 
  //  for (int i = 0; i < nb_swaps; i++) {
  // swap_link(g);
  // }    
  
  

  // 3. recompute asp
  //    recompute deterministic usp
  //    compute random usp
  asp_nodes2 = asp(g, src, dest_set, nb_dest, 
                   asp_links2, 
                   &nb_asp_nodes_part2, &nb_asp_links_part2);

  usp_nodes2 = deterministic_usp(g, src, dest_set, nb_dest, usp_links2, 
                                 &nb_usp_links_part2);
  nb_usp_nodes_part2 = nb_usp_links_part2 + 1;

  usp_nodes2_r = usp(g, src, dest_set, nb_dest, usp_links2_r, 
                                 &nb_usp_links_part2_r);
  nb_usp_nodes_part2_r = nb_usp_links_part2_r + 1;

  // 4. find differences between 1. and 3.
  // usp
  // ---
  //    'usp_nodes' <- union of 'usp_nodes' and 'usp_nodes2'
  nodes_union(usp_nodes, &usp_diff_nodes, 
              usp_nodes2, nb_usp_nodes_part2);
  // 1) 'usp_links_seen' <- union of 'usp_links_seen' and 'usp_links'
  // 2) 'usp_links_seen' <- union of 'usp_links_seen' and 'usp_links'
  links_union(usp_links_seen, &usp_diff_links,
              usp_links, nb_usp_links_part);
  usp_diff_links = 0;
  links_union(usp_links_seen, &usp_diff_links,
              usp_links2, nb_usp_links_part2);

  
  // asp
  // ---
  //    'asp_nodes' <- union of 'asp_nodes' and 'asp_nodes2'
  nodes_union(asp_nodes_seen, &asp_diff_nodes,
              asp_nodes, nb_asp_nodes_part);
  asp_diff_nodes = 0;
  nodes_union(asp_nodes_seen, &asp_diff_nodes,
              asp_nodes2, nb_asp_nodes_part2);
  // 1) 'asp_links_seen' <- union of 'usp_links_seen' and 'usp_links'
  // 2) 'asp_links_seen' <- union of 'usp_links_seen' and 'usp_links'
  links_union(asp_links_seen, &asp_diff_links,
              asp_links, nb_asp_links_part);
  asp_diff_links = 0;
  links_union(asp_links_seen, &asp_diff_links,
              asp_links2, nb_asp_links_part2);

  //clean mem after asp-step
  //prepare for 'random usp'-step
  clear_link_set(asp_links_seen, nb_asp_links_part + asp_diff_links);

  free_set(asp_nodes_seen, nb_asp_nodes_part + asp_diff_nodes);
  assert( (asp_nodes_seen = (struct bst_node **)malloc(sizeof(struct bst_node*))) != NULL );
  *asp_nodes_seen = NULL;

  //random usp
  nodes_union(asp_nodes_seen, &usp_diff_nodes_r,
              asp_nodes, nb_asp_nodes_part);
  usp_diff_nodes_r = 0;
  nodes_union(asp_nodes_seen, &usp_diff_nodes_r,
              usp_nodes2_r, nb_usp_nodes_part2_r);


  links_union(asp_links_seen, &usp_diff_links_r,
              asp_links, nb_asp_links_part);
  usp_diff_links_r = 0;
  links_union(asp_links_seen, &usp_diff_links_r,
              usp_links2_r, nb_usp_links_part2_r);

  // 5. print results!

  printf("%ld %ld %ld %ld %ld %ld %ld %ld %ld \n",
         nb_usp_nodes_part, 
         usp_diff_nodes, usp_diff_links,
         usp_diff_nodes_r, usp_diff_links_r,

         nb_asp_nodes_part, nb_asp_links_part,
         asp_diff_nodes, asp_diff_links);


  free_set(usp_nodes,  nb_usp_nodes_part + usp_diff_nodes);
  free_set(usp_nodes2, nb_usp_nodes_part2);
  free_set(usp_nodes2_r, nb_usp_nodes_part2_r);
  free_set(asp_nodes,  g->n);
  free_set(asp_nodes2, g->n);
  free_link_set(asp_links, g->m);
  free_link_set(usp_links, g->m);
  free_link_set(asp_links2, nb_asp_links_part2);
  free_link_set(usp_links2, nb_usp_links_part2);
  free_link_set(usp_links2_r, nb_usp_links_part2_r);
  free_link_set(usp_links_seen, nb_usp_links_part + usp_diff_links);
  free_link_set(asp_links_seen, nb_asp_links_part + usp_diff_links_r);
  free_set(asp_nodes_seen,      nb_asp_nodes_part + usp_diff_nodes_r);
  free_set(dest_set,nb_dest);
}


// computes probability of ASP/USP-changes
void main_proba_pl (int argc, char** argv) {
  long n,m,nb_dest,nb_step,nb_swaps;
  double p;

  if( argc != 7 ){
    fprintf(stderr, 
"Usage: %s nb_nodes exp max_links_for_that_exp nb_dest nb_swaps nb_step\n\
       Sorry for max_links_for_that_exp :) \n",argv[0]);
    exit(1);
  }

  sscanf(argv[1],"%ld",&n);
  sscanf(argv[2],"%lf",&p);
  sscanf(argv[3],"%ld",&m);
  sscanf(argv[4],"%ld",&nb_dest);
  sscanf(argv[5],"%ld",&nb_swaps);
  sscanf(argv[6],"%ld",&nb_step);

  //printf("# nodes_in_usp usp_diff_nodes usp_diff_links nodes_in_asp links_in_asp asp_diff_nodes asp_diff_links\n");
  // - what abour sizes of recalculated asp\usp?
  // - i suppose they are 'the same' as calculated sizes.
  
  printf("# nodes_in_usp   usp_diff_nodes   usp_diff_links r_usp_diff_nodes r_usp_diff_links nodes_in_asp links_in_asp asp_diff_nodes   asp_diff_links\n");

  for (int i = 0; i < nb_step; i++) {
    graph* g = gen_pl_graph(n, -p, m);
    run_proba_experiment(g, nb_dest, nb_swaps, nb_step);
    free_graph(g);
  }
  
}

void main_proba_er (int argc, char** argv) {
  long n,m,nb_dest,nb_step,nb_swaps;

  if( argc != 6 ){
    fprintf(stderr, 
            "Usage: %s nb_nodes nb_links nb_dest nb_swaps nb_step\n",argv[0]);
    exit(1);
  }

  sscanf(argv[1],"%ld",&n);
  sscanf(argv[2],"%ld",&m);
  sscanf(argv[3],"%ld",&nb_dest);
  sscanf(argv[4],"%ld",&nb_swaps);
  sscanf(argv[5],"%ld",&nb_step);

  printf("# nodes_in_usp usp_diff_nodes usp_diff_links nodes_in_asp links_in_asp asp_diff_nodes asp_diff_links\n");
  // - what abour sizes of recalculated asp\usp?
  // - i suppose they are 'the same' as calculated sizes.

  for (int i = 0; i < nb_step; i++) {
    graph* g = gen_er_graph(n, m);
    run_proba_experiment(g, nb_dest, nb_swaps, nb_step);
    free_graph(g);
  }

}

int main(int argc, char** argv) {
  //TODO: just use another seed here.
  srandom(time(NULL) + getpid());

  mallopt(M_TRIM_THRESHOLD,-1);
  mallopt(M_MMAP_MAX,0);

  if( argc < 2 ){
    fprintf(stderr, 
            "Usage: %s er|pl [bla..bla..bla]\n", argv[0]);
    exit(1);
  }
  
  if (strcmp(argv[1],"er") == 0) {
    main_proba_er(argc-1, argv+1);
  } else if (strcmp(argv[1],"pl") == 0) {
    main_proba_pl(argc-1, argv+1);
  } else {
    fprintf(stderr, 
            "Usage: %s er|pl [bla..bla..bla]\n",argv[0]);
    exit(1);
  }

  return (0);
}

